"""
datasets
--------

This module contains functions to get the paths to the datasets used in the
"""

# Imports
from numpy import ndarray, loadtxt
from importlib import resources
from typing import Tuple
import gzip


def get_ceman_counts():
    """
    Get path the GSE50499 [1]_ gene counts matrix.

    Returns
    -------
    pathlib.PosixPath
        Path to file

    References
    ----------
    .. [1] Kenny PJ, Zhou H, Kim M, Skariah G et al. MOV10 and FMRP regulate
           AGO2 association with microRNA recognition elements. Cell Rep 2014
           Dec 11;9(5):1729-1741. PMID: 25464849
    """

    with resources.path("infoml.data", "GSE50499_GEO_Ceman_counts.txt.gz") as f:
        datapath = f
    return datapath


def load_boston() -> Tuple[ndarray, ndarray, list]:
    """
    Load the Boston Housing dataset [1]_ for house porice prediction.

    Columns
    -------
    `CRIM`: per capita crime rate by town
    `ZN`: proportion of residential land zoned for lots over 25,000 sq.ft.
    `INDUS`: proportion of non-retail business acres per town
    `CHAS`: Charles River dummy variable (= 1 if tract bounds river; 0 otherwise)
    `NOX`: nitric oxides concentration (parts per 10 million)
    `RM`: average number of rooms per dwelling
    `AGE`: proportion of owner-occupied units built prior to 1940
    `DIS`: weighted distances to five Boston employment centres
    `RAD`: index of accessibility to radial highways
    `TAX`: full-value property-tax rate per $10,000
    `PTRATIO`: pupil-teacher ratio by town
    `B`: 1000(Bk - 0.63)^2 where Bk is the proportion of blacks by town
    `LSTAT`: % lower status of the population
    `MEDV`: Median value of owner-occupied homes in $1000's

    Returns
    -------
    Tuple[ndarray, ndarray, list]
        Tuple of (X, y, feature_names) where X is the input data, y is the
        target variable, and feature_names is the list of feature names.

    References
    ----------
    .. [1] Harrison, Jr., David, Rubinfeld, Daniel L. (1978/03)."Hedonic housing
           prices and the demand for clean air." Journal of Environmental
           Economics and Management 5(1): 81-102.
    """

    with resources.path("infoml.data", "boston-housing.txt.gz") as f:
        datapath = f
        with gzip.open(datapath, "rt") as f:
            rows = f.readlines()
            feature_names = rows[23].strip().split("\t")
            data = loadtxt(datapath, skiprows=24)

    X = data[:, :-1]
    y = data[:, -1]
    return X, y, feature_names



"""
infoml.utils
------------

This module contains utility functions for the infoml package.
Some of these functions are modified from the bmes package by Dr. Ahmet Sacan.
"""

# Imports from standard library
from tempfile import NamedTemporaryFile
from zipfile import ZipFile as zopen
from unicodedata import normalize
from itertools import islice
from shutil import copyfile
from pathlib import Path
import os, platform, re, sqlite3, subprocess

# Imports from third party packages
from pandas import DataFrame
from tqdm.auto import tqdm
from rich import print
import requests

# Imports from local source
from . import CONFIG


def ispc() -> bool:
    """Check if the current platform is Windows"""
    system = platform.system()
    return system == "Windows" or system.startswith("CYGWIN")


def isnonemptydir(path: Path | str) -> bool:
    """Check if a directory is non-empty"""
    path = Path(path)
    return path.is_dir() and len(os.listdir(path)) > 0


def isnonemptyfile(path: Path | str) -> bool:
    """Check if a file is non-empty"""
    path = Path(path)
    return path.is_file() and path.stat().st_size > 0


def tempfile(filename: str = "") -> Path:
    """Get a temporary file path"""
    if not filename:
        with NamedTemporaryFile() as f:
            filename = f.name
    return CONFIG.tempdir() / filename


def unzip(file: str | Path, dest: str | Path = ""):
    """
    Unzip a file

    Parameters
    ----------
    file : str | Path
        Path to the file to be unzipped
    dest : str | Path, optional
        Path to the destination directory, by default ''
    """

    # Convert file to Path
    file = Path(file)

    # Define destination directory
    if not dest:
        dest = CONFIG.tempdir() / Path(file).stem
        if not dest.exists():
            dest.mkdir(parents=True)
    else:
        dest = Path(dest)

    with zopen(file, "r") as zip:
        zip.extractall(dest)


def slugify(text: str, allow_unicode: bool = False) -> str:
    """
    Convert a string to a slug

    Based on Django's slugify function [1]_. This function converts a string to
    a slug. A slug is a string that contains only letters, numbers, underscores
    or hyphens. It is typically used to generate URL-friendly strings.

    Parameters
    ----------
    text : str
        The string to be sluggified
    allow_unicode : bool, optional
        Should unicode characters be permitted, by default False

    Returns
    -------
    str
        The sluggified string

    Raises
    ------
    AttributeError
        The input must be a string

    Examples
    --------
    >>> slugify("Jack & Jill like numbers 1,2,3 and 4 and silly characters ?%.$!/")
    'jack-jill-like-numbers-123-and-4-and-silly-characters'

    References
    ----------
    .. [1] Django. (n.d.). Django/text.py at main · Django/Django. GitHub.
           Retrieved January 2, 2023, from
           https://github.com/django/django/blob/main/django/utils/text.py
    """

    if not isinstance(text, str):
        raise AttributeError("The input must be a string")

    if allow_unicode:
        text = normalize("NFKC", text)
    else:
        text = normalize("NFKD", text).encode("ascii", "ignore").decode("ascii")

    text = re.sub(r"[^\w.\s-]", "", text).strip().lower()

    return re.sub(r"[-\s]+", "-", text)


def downloadurl(
    url: str,
    file: str | Path = CONFIG.tempdir(),
    overwrite: bool = False,
    progress: bool = True,
) -> Path:
    """
    Download and save file from a given URL

    Parameters
    ----------
    url : str
        The URL to download the file from
    file : str, optional
        Path to file (or directory) where downloaded file will be stored, by
        default the file will be saved to a temporary directory
    overwrite : bool, optional
        Should existing files be overwritten, by default False
    progress : bool, optional
        Should a progress bar be displayed, by default True

    Returns
    -------
    Path
        Path to downloaded file

    Raises
    ------
    FileNotFoundError
        If the file does not exist
    FileExistsError
        If the file already exists and overwrite is False
    ConnectionError
        If the URL is not valid

    Examples
    --------
    >>> downloadurl("https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_272x92dp.png")
    PosixPath('/tmp/googlelogo_color_272x92dp.png')
    """

    # Convert file to Path object
    file = Path(file)

    # If URL is not a remote address, assume it is a local file
    if not re.search(r"(http[s]?|ftp):\/\/", url):
        if not Path(url).exists():
            raise FileNotFoundError(f"File {url} does not exist")
        if not overwrite:
            if Path(file).exists():
                raise FileExistsError(f"File {file} already exists")
            copyfile(url, file)
            return Path(file)

    # Get file name from URL
    if file.is_dir():
        fname = slugify(url.split("?")[0].split("/")[-1])
        file = (file / fname).resolve()
    else:
        file = file.resolve()

    # Return file if it exists and overwrite is False
    if isnonemptyfile(file) and not overwrite:
        return file

    # Download the file
    # TODO: See if urllib.request.urlretrieve can be used instead
    r = requests.get(url, stream=True, allow_redirects=True, timeout=(3, 30))
    if r.status_code == 200:
        size = int(r.headers.get("content-length", 0))

        if progress:
            with open(file, "wb") as stream:
                with tqdm(
                    total=size,
                    unit="B",
                    unit_scale=True,
                    desc=file.name,
                    leave=False,
                    dynamic_ncols=True,
                    initial=0,
                ) as pbar:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:
                            stream.write(chunk)
                            pbar.update(len(chunk))

        else:
            with open(file, "wb") as stream:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        stream.write(chunk)

    elif r.status_code == 404:
        raise FileNotFoundError(f"File {url} does not exist")
    else:
        raise ConnectionError(f"{r.status_code}: Could not download file {url}")

    return file


class SQLite:
    """
    Wrapper for connecting to and querying SQLite databases

    Attributes
    ----------
    file : Path
        Path to SQLite database file
    conn : sqlite3.Connection
        Connection to SQLite database

    Methods
    -------
    execute(query: str, *args, **kwargs) -> sqlite3.Cursor
        Execute a query on the database
    select(query: str, *args, **kwargs) -> list[sqlite3.Row]
        Execute a query on the database and return the results as a DataFrame
    insert(table: str, data: dict, **kwargs) -> None
        Insert data into a table
    is_table(table: str) -> bool
        Check if a table exists in the database
    drop(table: str) -> None
        Drop a table from the database
    tables() -> dict
        Return a list of tables in the database and their schema
    close() -> None
        Close the connection to the database
    """

    def __init__(self, file: str | Path, quiet: bool = False, **kwargs) -> None:
        """Initialize SQLite class"""

        self.file = Path(file).resolve()
        if not self.file.suffix and not self.file.is_dir():
            self.file = self.file.with_suffix(".db")
        self.conn = sqlite3.connect(self.file, **kwargs)
        self.conn.row_factory = sqlite3.Row
        self.quiet = quiet
        if self.quiet:
            print(f"Connected to {self.file.name}")

    def __enter__(self):
        """Enter context manager"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit context manager"""
        self.conn.commit()
        self.conn.close()

    def __repr__(self) -> str:
        """Return string representation of SQLite class"""
        return f"{self.__class__.__name__}({self.file})"

    def __str__(self) -> str:
        """Return string representation of SQLite class"""
        return f"{self.__class__.__name__}({self.file})"

    def execute(self, query: str, *args, **kwargs) -> sqlite3.Cursor | DataFrame | None:
        """
        Execute a query

        Parameters
        ----------
        query : str
            Query to execute

        Returns
        -------
        sqlite3.Cursor
            Cursor object

        Raises
        ------
        AttributeError
            If the input is not a string

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> db.execute("CREATE TABLE IF NOT EXISTS test (id INTEGER PRIMARY KEY, name TEXT)")
        <sqlite3.Cursor object at 0x7f8b8c0b0e00>
        >>> db.execute("INSERT INTO test (name) VALUES ('John')")
        <sqlite3.Cursor object at 0x7f8b8c0b0e00>
        >>> db.execute("INSERT INTO test (name) VALUES ('Jane')")
        <sqlite3.Cursor object at 0x7f8b8c0b0e00>
        >>> db.close()
        """

        if not isinstance(query, str):
            raise AttributeError("The input must be a string")

        if query.upper().strip().startswith("SELECT"):
            return self.select(query, *args, **kwargs)

        try:
            cursor = self.conn.execute(query, *args, **kwargs)
            self.conn.commit()
            return cursor
        except sqlite3.Error as e:
            print(f"Error: {e}")
            return None

    def select(self, query: str, *args, **kwargs) -> DataFrame:
        """
        Execute a SELECT query and return the results as a DataFrame

        Parameters
        ----------
        query : str
            Query to execute

        Returns
        -------
        DataFrame
            Results of query

        Raises
        ------
        AttributeError
            If the input is not a string

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> db.select("SELECT * FROM test")
            id  name  age
            0  1  John   30
            1  2  Jane   25
        """

        if not isinstance(query, str):
            raise AttributeError("The input must be a string")

        try:
            cursor = self.conn.cursor()
            cursor.execute(query, *args, **kwargs)
            rows = cursor.fetchall()
        except sqlite3.Error as e:
            print(f"Error: {e}")
            return DataFrame()

        if len(rows) == 0:
            print("No results found")
            return DataFrame()
        else:
            df = DataFrame(rows)
            df.columns = [col[0] for col in cursor.description]  # type: ignore
            return df

    def insert(self, table: str, data: dict, **kwargs) -> None:
        """
        Insert data into a table

        Parameters
        ----------
        table : str
            Name of table
        data : dict
            Data to insert

        Raises
        ------
        AttributeError
            If the table does not exist
            If the dictionary values are not lists of the same length
            If a database column is missing from the dictionary

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> data = {'name': ['John', 'James', 'Rose', 'Jane'],
                    'age':  [30, 25, 60, 45]}
        >>> db.insert("test", data)
        >>> db.close()
        """

        # Check if table exists
        if not self.is_table(table):
            raise AttributeError(f"Table '{table}' does not exist")

        # Dictionary values must be lists of the same length
        if not all(len(v) == len(data[list(data.keys())[0]]) for v in data.values()):
            raise AttributeError("Dictionary values must be lists of the same length")
        nrows = len(data[list(data.keys())[0]])

        # Get column information
        info = self.select(f"PRAGMA table_info({table})")
        columns = info["name"].tolist()
        types = info["pk"].tolist()

        # Check if all columns are present (except for autoincrementing columns)
        for col, type in zip(columns, types):
            if col not in data and type == 0:
                raise AttributeError(f"Column '{col}' is missing")

        # Execute queries
        for _, vals in zip(range(nrows), zip(*list(data.values()))):
            cols = ", ".join(data.keys())
            subs = ", ".join("?" * len(data))
            query = f"INSERT INTO {table} ({cols}) VALUES ({subs})"
            self.execute(query, vals, **kwargs)

    def is_table(self, table: str) -> bool:
        """
        Check if a table exists in the database

        Parameters
        ----------
        table : str
            Name of table

        Returns
        -------
        bool
            True if table exists, False otherwise

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> db.is_table("test")
        True
        >>> db.is_table("foo")
        False
        """

        try:
            self.select(
                ("SELECT name FROM sqlite_master " "WHERE type='table' AND name=?"),
                (table,),
            )
        except sqlite3.OperationalError:
            return False

        return True

    def drop(self, table: str) -> None:
        """
        Drop a table from the database

        Parameters
        ----------
        table : str
            Name of table

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> db.drop("test")
        """

        self.execute(f"DROP TABLE {table}")

    def tables(self) -> dict:
        """
        List all tables in the database and their schema

        Returns
        -------
        dict
            Dictionary of table names and their schema

        Examples
        --------
        >>> db = SQLite("test.db")
        >>> db.tables()
        {'test': 'CREATE TABLE test (id INTEGER PRIMARY KEY, name TEXT)'}
        """

        tables = {}
        for row in self.select(
            ("SELECT name, sql FROM sqlite_master " "WHERE type='table'")
        ):
            tables[row["name"]] = row["sql"]  # type: ignore
        return tables

    def close(self) -> None:
        """Close the database connection"""

        self.conn.commit()
        self.conn.close()
        if self.quiet:
            print(f"Closed connection to {self.file.name}")


def iohead(file: str, n: int = 5) -> None:
    """
    Print the first n rows of a text file

    Parameters
    ----------
    file : str
        Name of file
    n : int, optional
        Number of rows to print
    """

    # Check if file exists
    if not isnonemptyfile(file):
        print("[red bold]ERROR:[/red bold] File not found.")
        return

    # Read and print file
    with open(file, "r") as f:
        for line in islice(f, n):
            if len(line) > 80:
                print(line[:80] + "...")
            else:
                print(line.rstrip())

    return


def system(
    command: str,
    stdout: str | None = None,
    stderr: str | None = None,
    quiet: bool = False,
    *args,
    **kwargs,
) -> str:
    """
    Run a system command; allows you to redirect stdout and stderr.

    Parameters
    ----------
    command : str
        Command to run
    stdout : str, optional
        File to redirect stdout to, by default ''
    stderr : str, optional
        File to redirect stderr to, by default ''
    quiet : bool, optional
        Suppress output, by default False

    Returns
    -------
    str
        Output from the command
    """

    if not quiet:
        print("[green bold]RUNNING:[/green bold] " + command)

    try:
        if stdout:
            command += f" > {stdout}"

        if stderr:
            command += f" 2> {stderr}"
        else:
            stderr = subprocess.STDOUT  # type: ignore

        out = subprocess.check_output(
            command,
            stderr=stderr,  # type: ignore
            shell=True,
            *args,
            **kwargs,  # type: ignore
        )
    except subprocess.CalledProcessError as E:
        out = E.output.decode()

    if not quiet:
        print(out.decode())

    return out.decode()


class Bidict:
    """
    A bidirectional dictionary

    Methods
    -------
    items
        Return a list of (key, value) tuples
    keys
        Return a list of keys
    values
        Return a list of values
    """

    def __init__(self, *args, **kwargs):
        """Class constructor"""
        self._dict = dict(*args, **kwargs)
        self._rev = dict((v, k) for k, v in self._dict.items())

    def __getitem__(self, key):
        """Get item from dictionary"""
        return self._dict[key]

    def __setitem__(self, key, value):
        """Set item in dictionary"""
        self._dict[key] = value
        self._rev[value] = key

    def __delitem__(self, key):
        """Delete item from dictionary"""
        del self._rev[self._dict[key]]
        del self._dict[key]

    def __iter__(self):
        """Iterate over dictionary"""
        return iter(self._dict)

    def __len__(self):
        """Return length of dictionary"""
        return len(self._dict)

    def __repr__(self):
        """Return string representation of dictionary"""
        return repr(self._dict)

    def __str__(self):
        """Return string representation of dictionary"""
        return str(self._dict)

    def __contains__(self, key):
        """Check if key is in dictionary"""
        return key in self._dict

    def items(self):
        """Return a list of (key, value) tuples"""
        return self._dict.items()

    def keys(self):
        """Return a list of keys"""
        return self._dict.keys()

    def values(self):
        """Return a list of values"""
        return self._dict.values()

    @property
    def rev(self):
        """Return the reverse dictionary"""
        return self._rev


if __name__ == "__main__":
    print("This module is not intended to be run directly.")
else:
    # Define module I/O
    __all__ = [
        "ispc",
        "isnonemptydir",
        "isnonemptyfile",
        "tempfile",  # --
        "unzip",  # --
        "slugify",
        "downloadurl",  # --
        "SQLite",  # --
        "iohead",  # --
        "system",  # --
        "Bidict",
    ]
    __all__ += [m for m in dir() if m.startswith("__")]

    def __dir__():
        """Override default dir() behavior"""
        return __all__

    def __getattr__(name):
        """Override default getattr() behavior"""
        if name not in __all__:
            raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
        return globals()[name]



"""
infoml
------

`infoml` is a Python package for perfoming bioinformatics analysis and machine
learning tasks on genomic and bioimaging data.
"""

# Imports from standard library
from importlib.metadata import version
from socket import gethostname
from pathlib import Path
from rich import print
import sys, tempfile


class config:
    """
    Configuration class for infoml package

    Methods
    -------
    cache : pathlib.PosixPath
        Path to the cache directory. This will be the default location for
        storing cached data.
    datadir : pathlib.PosixPath
        Path to the data directory. This will be the default location for
        storing downloaded data.
    tempdir : pathlib.PosixPath
        Path to the temporary directory. This will be the default location for
        storing temporary files.

    Examples
    --------
    >>> from infoml import CONFIG
    >>> CONFIG.DATADIR
    PosixPath('/home/kabil/.data/infoml')
    >>> # Change the default data directory
    >>> CONFIG.DATADIR = "./data"
    """

    __cache = Path.home() / ".cache" / "infoml"
    __datadir = Path.home() / ".data" / "infoml"
    __tempdir = Path(tempfile.gettempdir()) / "infoml"

    def __init__(self):
        """Initialize the configuration class"""

        if not self.__cache.exists():
            self.__cache.mkdir(parents=True)

        if not self.__datadir.exists():
            self.__datadir.mkdir(parents=True)

        if not self.__tempdir.exists():
            self.__tempdir.mkdir(parents=True)

    def cache(self, new_path: Path | str = "") -> Path:
        """Cache directory"""

        if new_path:
            self.__cache = Path(new_path)

            if not self.__cache.exists():
                self.__cache.mkdir(parents=True)

        return self.__cache

    def datadir(self, new_path: Path | str = "") -> Path:
        """Data directory"""

        if new_path:
            self.__datadir = Path(new_path)

            if not self.__datadir.exists():
                self.__datadir.mkdir(parents=True)

        return self.__datadir

    def tempdir(self, new_path: Path | str = "") -> Path:
        """Temporary directory"""

        if new_path:
            self.__tempdir = Path(tempfile.gettempdir()) / new_path

            if not self.__tempdir.exists():
                self.__tempdir.mkdir(parents=True)

        return self.__tempdir

    def sysinfo(self) -> dict:
        """Get System Information"""
        return {
            "host_name": gethostname(),
            "operating_system": sys.platform,
            "python_version": sys.version,
            "python_executable": sys.executable,
            "infoml_version": version("infoml"),
            "infoml_path": Path(__file__).parent.absolute(),
        }

    def __str__(self):
        """String representation of the class"""
        return str(
            dict(
                cache=self.cache(),
                datadir=self.datadir(),
                tempdir=self.tempdir(),
            )
        )

    def __repr__(self):
        """Representation of the class"""
        return str(self)


# Export modules
CONFIG = config()
from . import datasets
from . import utils
from . import pkg
from . import viz
from . import binf


# Deine package version
__version__ = version(__name__)


if __name__ == "__main__":
    print("This module is not intended to be run directly.")
else:
    # Define module I/O
    __all__ = [
        "CONFIG",
        "datasets",
        "utils",
        "pkg",
        "viz",
        "binf",
    ]
    __all__ += [m for m in dir() if m.startswith("__")]

    def __dir__():
        """Override default dir() behavior"""
        return __all__

    def __getattr__(name):
        """Override default getattr() behavior"""
        if name not in __all__:
            raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
        return globals()[name]



"""
infoml.binf.data
----------------

This module contains functions for downloading data from bioinformatics
repositiories like Gene Expression Omnibus and the Curated Microarray Database.
"""

# Imports from standard library
from pathlib import Path
from typing import Union
import warnings
import pickle
import json
import re
import os

# Imports from third party packages
from GEOparse.GEOTypes import GPL, GSE
from GEOparse import get_GEO
from tqdm.auto import tqdm
from rich import print
import pandas as pd
import numpy as np

# Imports from local source
from ..utils import CONFIG, downloadurl


# Suppress DtypeWarning from GEOparse
warnings.filterwarnings(
    action="ignore", category=pd.errors.DtypeWarning, module="GEOparse"
)


def geodlparse(
    acc: str,
    datadir: str | Path = "",
    silent: bool = True,
    make_dir: bool = False,
    cache: bool = False,
) -> GSE | GPL:  # type: ignore
    """
    Download, parse and cache data from GEO.
    This fuction only downloads GSE and GPL data.

    Parameters
    ----------
    acc : str
        GEO accession
    datadir : str | Path, optional
        Directory for storing downloaded data, will default to a
        temporary directory if not specified
    silent : bool, optional
        Whether to suppress output, by default False
    make_dir : bool, optional
        Whether to make the directory if it does not exist,
        by default False
    cache : bool, optional
        Whether to cache the data, by default False

    Returns
    -------
    GPL | GSE
        Parsed GEO data
    """

    # Check inputs
    acc = acc.upper()
    assert isinstance(acc, str), "acc must be a string"
    assert acc.startswith("GSE") or acc.startswith(
        "GPL"
    ), "acc must be a GSE or GPL accession"

    assert isinstance(datadir, str) or isinstance(
        datadir, Path
    ), "datadir must be a string or pathlib.Path object"
    if datadir == "":
        # Use a temporary directory
        datadir = CONFIG.tempdir()
    elif not os.path.exists(datadir):
        if make_dir:
            os.makedirs(datadir)
        else:
            raise ValueError("Directory does not exist")
    if isinstance(datadir, str):
        datadir = Path(datadir)

    assert isinstance(silent, bool), "silent must be a boolean"
    assert isinstance(make_dir, bool), "make_dir must be a boolean"

    # Define file names
    geofile = datadir.joinpath(
        f"{acc}.txt" if acc[:3] == "GPL" else f"{acc}_family.soft.gz"
    ).resolve()
    cachefile = (CONFIG.cache() / f"{acc}.cache").resolve()

    # Load cached data if it exists
    if cachefile.is_file():
        try:
            if not silent:
                print(f"Loading cached data for {acc}")
            with open(cachefile, "rb") as f:
                return pickle.load(f)
        except Exception as E:
            print(
                f"[bold red]Error loading cached data[/bold red]", f"\n\n{E}", sep=" "
            )

    # Download, parse and cache data
    else:
        try:
            # Parse already downloaded data
            if os.path.isfile(geofile):
                if not silent:
                    print(f"Parsing {acc}")
                geodata = get_GEO(filepath=geofile.__str__(), silent=silent)

            # Download and parse data
            else:
                if not silent:
                    print(f"Downloading and parsing {acc}")
                geodata = get_GEO(acc, destdir=datadir, silent=silent)  # type: ignore

            # Cache data
            if cache:
                with open(cachefile, "wb") as handle:
                    pickle.dump(geodata, file=handle)

            return geodata  # type: ignore

        except OSError as E:
            print(
                "[bold red]Error[/bold red]: It seems you've entered",
                f"an invalid accession number.\n\n{E}",
                sep=" ",
            )

        except Exception as E:
            print(
                "[bold red]Error[/bold red]: Something went wrong.", f"\n\n{E}", sep=" "
            )


class CuMiDa:
    """
    Class for loading datasets from the Curated Microarray Database
    hosted by SBCB (sbcb.inf.ufrgs.br)

    Attributes
    ----------
    INDEX : str
        Path to the JSON file containing an index of all datasets in the
        database.
    BASEURL : str
        Base URL for downloading datasets from CuMiDa.
    index : pd.DataFrame
        Index of all datasets available from CuMiDa.
    datadir : str | Path
        Directory for storing downloaded data.
    gse_dir : str | Path
        Directory for storing downloaded GSE data.
    gpl_dir : str | Path
        Directory for storing downloaded GPL data.

    Methods
    -------
    download(acc: str, datadir: str | Path='', silent: bool=False,
             make_dir: bool=False)
        Download a dataset from CuMiDa.
    load(dataset: tuple)
        Load a specified dataset, along with gene annotations from its GPL.
    """

    INDEX = (
        "https://gist.githubusercontent.com/Kabilan108/"
        "3d11266abdd3c237d359dd7c11a40871/raw/"
        "ff2af81ae70afaba99233400f9d79e30eb40942e/cumida.json"
    )
    BASEURL = "https://sbcb.inf.ufrgs.br"

    def __init__(self, datadir: Union[str, Path] = "") -> None:
        """
        Initialize the CuMiDa class.

        Parameters
        ----------
        datadir : str or Path, optional
            Directory for storing downloaded data, will default to the
            project's data directory, utils.datadir() if not specified
        """

        # Check inputs
        assert isinstance(datadir, str) or isinstance(
            datadir, Path
        ), "datapath must be a string or PosixPath"
        if datadir == "":
            datadir = CONFIG.datadir()
        if isinstance(datadir, str):
            datadir = Path(datadir)
        if not os.path.exists:
            os.makedirs(datadir)
        self.datadir = datadir.resolve()

        # Retrieve the index of datasets
        self._makeindex()

        # Create subdirectory for gene expression matrices and platforms
        self.gse_dir = self.datadir / "GSE"
        self.gpl_dir = self.datadir / "GPL"
        if not os.path.exists(self.gse_dir):
            os.makedirs(self.gse_dir)
        if not os.path.exists(self.gpl_dir):
            os.makedirs(self.gpl_dir)

    def _makeindex(self) -> None:
        """
        Create a dataframe containing the index of all datasets
        available from CuMiDa.
        """

        # Download the index
        file = (self.datadir / "datasets.json").resolve().__str__()
        downloadurl(self.INDEX, file, progress=False)

        # Load the dataset index
        with open(self.datadir / "datasets.json", "rb") as f:
            self.index = pd.DataFrame(json.load(f))

        # Clean up the columns
        self.index["ID"] = "GSE" + self.index["gse"].astype(str)
        self.index["Platform"] = "GPL" + self.index["platform"].astype(str)
        self.index["URL"] = (
            self.BASEURL + self.index["downloads"].apply(pd.Series)["csv"]
        )
        self.index = (
            self.index.rename(
                columns={
                    "type": "Type",
                    "classes": "Classes",
                    "samples": "Samples",
                    "genes": "Genes",
                    "manufacturer": "Manufacturer",
                }
            )[
                [
                    "ID",
                    "Platform",
                    "Manufacturer",
                    "Type",
                    "Classes",
                    "Samples",
                    "Genes",
                    "URL",
                ]
            ]
            .sort_values(by=["Platform", "Type"])
            .set_index(["ID", "Type"])
        )
        self._downloads = self.index["URL"].to_dict()

    def download(self, selected: pd.DataFrame | tuple | list) -> None:
        """
        Download selected datasets from CuMiDa.

        Parameters
        ----------
        selected : pd.DataFrame | tuple
            A subset of `self.index` containing the datasets to download.
            Or a tuple of (ID, Type) for a single dataset or a list of tuples.

        Returns
        -------
        None
        """

        try:
            if isinstance(selected, pd.DataFrame):
                urls = [self._downloads[x] for x in selected.index]
                self._selected = selected.index.to_list()
            elif isinstance(selected, tuple):
                urls = [self._downloads[selected]]
                self._selected = [selected]
            elif isinstance(selected, list):
                urls = [self._downloads[x] for x in selected]
                self._selected = selected
            else:
                raise TypeError("selected must be a DataFrame or tuple")
        except KeyError:
            raise KeyError("Dataset not found in CuMiDa index")

        # Download the GSE matrices from CuMiDa
        self.file_paths = [
            self.gse_dir / re.search(r"\w+\.csv", x)[0] for x in urls  # type: ignore
        ]
        if len(self.file_paths) > 1:
            with tqdm(total=len(urls), desc="Downloading GSEs") as pbar:
                for url, file in zip(urls, self.file_paths):
                    downloadurl(url, file.__str__(), progress=False)
                    pbar.update(1)
        else:
            downloadurl(urls[0], self.file_paths[0].__str__(), progress=False)

        # Download the GPLs from GEO
        self._gpl_accs = np.unique(
            [self.index.loc[x]["Platform"] for x in self._selected]
        )
        self._gpls = dict()
        with tqdm(total=len(self._gpl_accs), desc="Downloading GPLs") as pbar:
            for acc in self._gpl_accs:
                self._gpls[acc] = geodlparse(acc, self.gpl_dir.__str__(), silent=True)
                pbar.update(1)

    def load(self, dataset: tuple, probe_ids: bool = False) -> pd.DataFrame:
        """
        Load a specified dataset.

        Parameters
        ----------
        dataset : tuple
            A tuple of (ID, Type) for a single dataset.
        probe_ids : bool, optional
            If true, return the probe IDs instead of the GenBank Accessions.

        Returns
        -------
        gse : pd.DataFrame
        """

        # Check inputs
        assert isinstance(dataset, tuple), "dataset must be a tuple"
        assert len(dataset) == 2, "dataset must be a tuple of (ID, Type)"
        assert dataset in self.index.index, "dataset not found in CuMiDa index"

        # Load the GSE
        path = self.gse_dir / f"{'_'.join(dataset[::-1])}.csv"
        gse = pd.read_csv(path).set_index(["samples", "type"])

        # Rename the columns
        if not probe_ids:
            try:
                # Rename GSE columns with GenBank IDs where possible
                gpl = self._gpls[self.index.loc[dataset]["Platform"]].table
                gpl["GB_ACC"] = gpl["GB_ACC"].fillna(gpl["ID"])
                gse.columns = gse.columns.map(gpl.set_index("ID")["GB_ACC"])

                # Add numeric suffices to duplicate column names
                idx = gse.columns.to_series().groupby(level=0).transform("cumcount")
                gse.columns = gse.columns + "." + idx.astype(str)

                # Sort columns alphabetically
                gse = gse.reindex(sorted(gse.columns), axis=1)
            except KeyError as E:
                print(f"No GenBank IDs found for {dataset[0]}")
                print(E)

        return gse

    def __repr__(self) -> str:
        """Return a string representation of the CuMiDa class"""

        return f"CuMiDa(datadir={self.datadir})"

    def __str__(self) -> str:
        """String representation of the CuMiDa class"""

        return self.__repr__()



"""
infoml.themes
-----------

This module contains themes for plotly and matplotlib (seaborn included).
"""

from abc import ABC, abstractmethod
import matplotlib.pyplot as plt


class FigureTheme(ABC):
    """
    Abstract base clas for figure themes.
    """

    @abstractmethod
    def apply(self, fig):
        """
        Apply the theme to a figure.

        Parameters
        ----------
        fig : object
            A figure object for the corresponding plotting library.

        This method should be implemented by concrete theme classes to apply
        the specific aesthetic defined by the theme to a figure.
        """
        pass


# class MatplotlibTheme(FigureTheme):
#     """
#     A theme for matplotlib figures.
#     """

#     def __init__(self, style="graphpaper"):
#         """
#         Initialize a matplotlib theme.

#         Parameters
#         ----------
#         style : str, optional
#             The style to use for the theme. The default is "graphpaper".
#         """
#         self._style = style

#     def apply(self, fig):
#         """
#         Apply the theme to a figure.

#         Parameters
#         ----------
#         fig : object
#             A figure object for the corresponding plotting library.

#         Returns
#         -------
#         None.

#         """
#         # Check if style is built-in
#         if self._style in plt.style.available:
#             plt.style.use(self._style)
#         else:
#             raise ValueError(f"Style {self._style} is not available.")


class MplTheme(FigureTheme):
    """
    A theme for matplotlib figures.
    """

    def __init__(self, style="graphpaper"):
        self._style = style

    def apply(self, fig, axis, **kwargs):
        # Check if style is built-in
        if self._style in plt.style.available:
            plt.style.use(self._style)

        elif self._style == "graphpaper":
            # Read arguments
            gridcolor = kwargs.get("gridcolor", "black")
            majorgridwidth = kwargs.get("majorgridwidth", 0.5)
            minorgridwidth = kwargs.get("minorgridwidth", 0.5)
            majorgridalpha = kwargs.get("majorgridalpha", 0.5)
            minorgridalpha = kwargs.get("minorgridalpha", 0.2)

            # Set style
            axis.spines[["top", "right"]].set_visible(False)
            axis.spines[["left", "bottom"]].set_linewidth(1.2)
            axis.xaxis.set_ticks_position("bottom")
            axis.yaxis.set_ticks_position("left")
            axis.set_xlabel("Time (s)")
            axis.minorticks_on()
            axis.grid(
                which="major",
                linestyle="-",
                linewidth=majorgridwidth,
                color=gridcolor,
                alpha=majorgridalpha,
            )
            axis.grid(
                which="minor",
                linestyle=":",
                linewidth=minorgridwidth,
                color=gridcolor,
                alpha=minorgridalpha,
            )

    def __enter__(self, *args, **kwargs):
        self._fig, self._axis = plt.subplots(*args, **kwargs)
        self.apply(self._fig, self._axis)
        return self._axis

    def __exit__(self, *args, **kwargs):
        plt.show()


class PlotlyTheme(FigureTheme):
    """
    A theme for plotly figures.
    """

    def __init__(self, style="plotly"):
        """
        Initialize a plotly theme.

        Parameters
        ----------
        style : str, optional
            The style to use for the theme. The default is "plotly".
        """
        self._style = style

    def apply(self, fig):
        """
        Apply the theme to a figure.

        Parameters
        ----------
        fig : object
            A figure object for the corresponding plotting library.

        Returns
        -------
        None.

        """
        fig.update_layout(template=self._style)


if __name__ == "__main__":
    print("This module is not intended to be run directly.")
else:
    # Define module I/O
    __all__ = []
    __all__ += [m for m in dir() if m.startswith("__")]

    def __dir__():
        """Override dir() to only show public attributes."""
        return __all__

    def __getattr__(name):
        """Override getattr() to only show public attributes."""
        if name in __all__:
            return globals()[name]
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
infoml.viz
----------

This sub-package contains functions for visualizing data.
"""

from . import themes


if __name__ == "__main__":
    print("This module is not intended to be run directly.")
else:
    # Define module I/O
    __all__ = [
        "themes",
    ]
    __all__ += [m for m in dir() if m.startswith("__")]

    def __dir__():
        """Override dir() to only show public attributes."""
        return __all__

    def __getattr__(name):
        """Override getattr() to only show public attributes."""
        if name in __all__:
            return globals()[name]
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""
infoml.pkg
----------

This module contains functions for managing - install, uninstall, check - 
python packages with both `pip` and `conda`.
"""

# Imports from standard library
import sys

# Imports from local source
from .utils import system


def check_package(importname: str) -> bool:
    """
    Check if a package is installed.

    Parameters
    ----------
    importname : str
        Name used when importing the package

    Returns
    -------
    bool
        True if package is installed, False otherwise
    """

    try:
        __import__(importname)
    except ImportError:
        return False

    return True


def pipinstall(
    packagename: str,
    importname: str = "",
    version: str = "",
    reinstall: bool = False,
) -> None:
    """
    Install a python package with `pip`.

    Modified from the `bmes` package by Ahmet Sacan.

    Parameters
    ----------
    packagename : str, optional
        Name used for installing the package
    importname : str
        Name used when importing the package, by default None
        In most cases, this is identical to `packagename` but both must be
        provided when they are different.
    version : str, optional
        Version of the package to install, by default None
    reinstall : bool, optional
        Should the package be reinstalled, by default False

    Examples
    --------
    >>> pipinstall('numpy');
    >>> pipinstall('pandas', version='1.0.0');
    >>> pipinstall('biopython', 'Bio')
    """

    # Function for construction commands
    def __construct_cmd():
        if version is None:
            cmd = f"{sys.executable} -m pip install -U {packagename}"
        else:
            cmd = f"{sys.executable} -m pip install -U {packagename}=={version}"
        return cmd

    # Set the import name
    importname = importname or packagename

    # Check if the package is already installed
    if check_package(importname):
        if reinstall:
            print(f"Reinstalling {packagename}...")
            pipuninstall(packagename, importname)
        else:
            print(f"{packagename} is already installed.")
            return

    # Install the package
    out = system(__construct_cmd())
    known_errors = [
        "ERROR: Could not install packages",
        "Consider using the `--user` option",
    ]

    # Try again with `--user` option
    if known_errors[0] in str(out) and known_errors[1] in str(out):
        out = system(__construct_cmd() + " --user")

    # Check if package was installed
    if check_package(importname):
        print(f"{packagename} is installed.")
    else:
        print(f"{packagename} could not be installed.")
        print(out)

    return


def pipuninstall(packagename: str, importname: str = "") -> None:
    """
    Uninstall a python package with `pip`.

    Parameters
    ----------
    packagename : str
        Name used for installing the package
    importname : str, optional
        Name used when importing the package, by default None
        In most cases, this is identical to `packagename` but both must be
        provided when they are different.

    Examples
    --------
    >>> pipuninstall('numpy');
    """

    # Set the import name
    importname = importname or packagename

    # Check if the package is installed
    if not check_package(importname):
        print(f"{packagename} is not installed.")
        return

    # Uninstall the package
    out = system(f"{sys.executable} -m pip uninstall -y {packagename}")

    # Check if package was uninstalled
    if not check_package(importname):
        print(f"{packagename} is uninstalled.")
    else:
        print(f"{packagename} could not be uninstalled.")
        print(out)

    return


if __name__ == "__main__":
    print("This module is not intended to be run directly.")
else:
    # Define module I/O
    __all__ = [
        "check_package",
        "pipinstall",
        "pipuninstall",
    ]
    __all__ += [m for m in dir() if m.startswith("__")]

    def __dir__():
        """Override default dir() behavior"""
        return __all__

    def __getattr__(name):
        """Override default getattr() behavior"""
        if name not in __all__:
            raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
        return globals()[name]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['infoml', 'infoml.binf', 'infoml.data', 'infoml.viz']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.80,<2.0',
 'geoparse>=2.0.3,<3.0.0',
 'numpy>=1.22',
 'pandas>=1.5.2,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=13.0.0,<14.0.0',
 'scikit-learn>=1.2.0,<2.0.0',
 'scipy>=1.9.3,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'infoml',
    'version': '0.8.0',
    'description': 'Python package for bioinformatics analysis and machine learning.',
    'long_description': '<h1 align="center">infoml</h1>\n\n<p align="center">\n<a href="https://github.com/psf/black/blob/main/LICENSE">\n    <img alt="License: MIT" src="https://img.shields.io/github/license/Kabilan108/infoml">\n</a>\n<a href="https://github.com/psf/black">\n    <img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg">\n</a>\n<a href="https://codecov.io/gh/Kabilan108/infoml" >\n    <img src="https://codecov.io/gh/Kabilan108/infoml/branch/main/graph/badge.svg?token=38Y14PAAQQ"/>\n</a>\n<a href="https://github.com/Kabilan108/infoml/actions/workflows/CI-CD.yml">\n    <img alt="Test Status" src="https://github.com/Kabilan108/infoml/actions/workflows/CI-CD.yml/badge.svg?branch=main">\n</a>\n<a href="https://wakatime.com/badge/user/6a085912-85f1-47f5-acc7-c7f5ac1110ab/project/51ef67da-2e82-431e-9dc9-9bdd6d1c3d48">\n    <img\n    src="https://wakatime.com/badge/user/6a085912-85f1-47f5-acc7-c7f5ac1110ab/project/51ef67da-2e82-431e-9dc9-9bdd6d1c3d48.svg"\n    alt="wakatime">\n</a>\n</p>\n\nPython package for bioinformatics analysis and machine learning.\n\n## Installation\n\n```bash\npip install infoml\n```\n\n## Usage\n\n- TODO\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note\nthat this project is released with a Code of Conduct. By contributing to this\nproject, you agree to abide by its terms.\n\n## License\n\n`infoml` was created by Tony Kabilan Okeke. It is licensed under the terms of\nthe MIT license.\n\n## Credits\n\n`infoml` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/)\nand the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n',
    'author': 'Tony Kabilan Okeke',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
